//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include <algorithm>
#include "nProcess.h"
#include "nCore.h"
#include "nSGNS.h"
#include "nMaths.h"
#include "nSpreading.h"
#include "fftw_interface.h"
#include "nFFT_pas.h"
#include "nFillFFT.h"
#include "nRemoveBits.h"
#include "nOutput.h"

//namespace {

static const char UnitName [] = "nProcess";


void Fill_Last_from_SGNS()
{
	int sa_i;
	double temp_result;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		temp_result = SGNSFFT.Last_Results[Global.channel][sa_i];
		FFT_last_result[sa_i] = temp_result;
		FFT_raw_result[sa_i] += temp_result;
	}
}


void Fill_Last_from_History()
{
	int sa_i;
	double temp_result;

	for (sa_i = 0; sa_i <= Global.FFT.length_half;  ++sa_i)
	{
		temp_result = history.Last_Results[Global.channel][sa_i];
		FFT_last_result[sa_i] = temp_result;
		FFT_raw_result[sa_i] += temp_result;
	}
}


void Fill_History_from_Last()
{
	int sa_i;
	double temp_result;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		temp_result = FFT_last_result[sa_i];
		history.Last_Results[Global.channel][sa_i] = temp_result;
		temp_result = OneOver[process.Analyses_Performed - 1] * (FFT_raw_result[sa_i] - temp_result);
		history.DATA_results[Global.channel][sa_i] = history.DATA_results[Global.channel][sa_i] + temp_result;
		SGNSFFT.FFT_results_short[sa_i] = temp_result;
	}
}


void Fill_SGNSFFT_from_Last()
{
	int sa_i;
	double temp_result;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		temp_result = FFT_last_result[sa_i];
		SGNSFFT.Last_Results[Global.channel][sa_i] = temp_result;
		SGNSFFT.FFT_results_long[sa_i] = FFT_raw_result[sa_i] - temp_result;
	}
}


void Fill_Last_with_Zero()
{
	int sa_i;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		FFT_last_result[sa_i] = 0;
		FFT_skew_result[sa_i] = 0;
	}
}


void Zero_FFT_Raw_Results()
{
	int sa_i;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		FFT_raw_result[sa_i] = 0;
	}
}


void Add_To_BTRD_History()
{
	int sa_i;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		history.BTRD_results[Global.channel][sa_i] = history.BTRD_results[Global.channel][sa_i] + (FFT_raw_result[sa_i] * OneOver[process.Analyses_Performed]);
	}
}


void Add_To_CORR_History()
{
	int sa_i;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		history.CORR_results[Global.channel][sa_i] += (FFT_raw_result[sa_i] * OneOver[process.Analyses_Performed]);
	}
}


void Add_to_DATA_LongDist()
{
	int sa_i;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		LongDist.DATA_results[Global.channel][sa_i] += (FFT_raw_result[sa_i] * OneOver[process.Analyses_Performed]);
	}
}


void Add_to_BTRD_LongDist()
{
	int sa_i;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		LongDist.BTRD_results[Global.channel][sa_i] += (FFT_raw_result[sa_i] * OneOver[process.Analyses_Performed]);
	}
}


void Add_to_CORR_LongDist()
{
	int sa_i;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		LongDist.CORR_results[Global.channel][sa_i] += (FFT_raw_result[sa_i] * OneOver[process.Analyses_Performed]);
	}
}


void Post_Process_FFT_Array_to_Last()
{
	int sa_i;
	double sa_x;

	for (sa_i = 0; sa_i <= Global.FFT.length_half; ++sa_i)
	{
		/*with FFT_Array.DComplex[sa_i] do*/
		tDComplex& c = FFT_Array.DComplex[sa_i];
		sa_x = (c.Re * c.Re + c.Im * c.Im);
		FFT_last_result[sa_i] = sa_x;
		FFT_raw_result[sa_i] += sa_x;
	}
}


void Post_Process_FFT_Results()
{
	int sc_i;
	double sc_x;

	for (sc_i = 0; sc_i <= Global.FFT.length_half; ++sc_i)
	{
		/*with FFT_Array.DComplex[sc_i] do*/
		sc_x = nsqr(FFT_Array.DComplex[sc_i].Re) + nsqr(FFT_Array.DComplex[sc_i].Im);
		FFT_last_result[sc_i] = sc_x;
		FFT_raw_result[sc_i] += sc_x;
		FFT_skew_result[sc_i] = nroot(sc_x) * Skewing_Gain[sc_i << Global.FFT.bit_shift_from_max];
	}
}


void Post_Process_FFT_Results_No_Raw()
{
	int sc_i;
	double sc_x;

	for (sc_i = spreading.Bins.Lo[Global.FFT.bit_length]; sc_i <= spreading.Bins.hi[Global.FFT.bit_length] + 1; ++sc_i)
	{
		/*with FFT_Array.DComplex[sc_i] do*/
		sc_x = nsqr(FFT_Array.DComplex[sc_i].Re) + nsqr(FFT_Array.DComplex[sc_i].Im);
		FFT_skew_result[sc_i] = nroot(sc_x) * Skewing_Gain[sc_i << Global.FFT.bit_shift_from_max];
	}
}


void Add_to_Histogram()
{
	int sa_i, sa_j;

	for (sa_j = 0; sa_j < Global.channels; ++sa_j)
		for (sa_i = 0; sa_i < BlockData.Size.This; ++sa_i)
		{
			++ history.Histogram_DATA[round(BlockData.Pointer[2][sa_j][sa_i] * history.Histogram_Multiplier) + history.Histogram_Offset];
			++ history.Histogram_BTRD[round(BlockData.BTRD[sa_j][sa_i] * history.Histogram_Multiplier) + history.Histogram_Offset];
			++ history.Histogram_CORR[round(BlockData.CORR[sa_j][sa_i] * history.Histogram_Multiplier) + history.Histogram_Offset];
		}
}


int min_neg_codec_block_start;
int max_pos_codec_block_start;
int this_analysis_block_start;
double this_fft_underlap_length;
double (* Fill_FFT_Proc)() = NULL;
void (* Add_to_LongDist_Proc)() = NULL;


void Set_FFT_Variables()
{
	if (! FFTW_Initialised() && SetFFTLength(Global.FFT.length_half) == 0)
	{
		lossyWAVerror("Fatal FFT error!", 0x88);
	}

	min_neg_codec_block_start = -(BlockData.Size.Last + BlockData.Size.Prev);
	max_pos_codec_block_start = BlockData.Size.This + BlockData.Size.Next - Global.FFT.length;
	this_analysis_block_start = process.actual_analysis_blocks_start[Global.FFT.bit_length];
	this_fft_underlap_length = process.FFT_underlap_length[Global.FFT.bit_length];
}

void Process_Data_to_LongDist()
{
	int this_analysis_block_number;
	Zero_FFT_Raw_Results();
	process.Analyses_Performed = 0;

	for (this_analysis_block_number = 0; this_analysis_block_number < process.analysis_blocks[Global.FFT.bit_length]; ++this_analysis_block_number)
	{
		FillFFT.actual_analysis_block_start = int(this_analysis_block_start + this_analysis_block_number * this_fft_underlap_length);
		FillFFT.actual_analysis_block_start = std::min(std::max(FillFFT.actual_analysis_block_start, min_neg_codec_block_start), max_pos_codec_block_start);
		++ process.Analyses_Performed;

		if (Fill_FFT_Proc != 0)
		{
			if (FFTW_Initialised())
			{
				FFTW.Execute(FFTW.Plans[Global.FFT.bit_length]);
			}
			else
			{
				FFT_DIT_Real();
			}

			Post_Process_FFT_Array_to_Last();
		}
	}

	Add_to_LongDist_Proc();
}


// } // anonymous namespace


// EXPORT


void Process_This_Codec_Block()
{
	int this_channel;
	int this_analysis_number;
	int this_analysis_block_number;
	int codec_block_dependent_bits_to_remove;
	int local_channels;
	//double Spreading_Old;
	//double Spreading_New;
	double Spreading_result;
	double bits_removed_this_codec_block;
	//double temp_result;
	//int sa_i;
	//int sa_j;
	//int sa_k;
	//int temp_btr;
	int Spreading_Used;
	channel_data_rec channel_data [ 8 ];


	codec_block_dependent_bits_to_remove = Global.bits_per_sample;
	bits_removed_this_codec_block = 0;

	if (parameters.midside && (Global.channels == 2))
	{
		local_channels = 4;
	}
	else
	{
		local_channels = Global.channels;
	}

	for (this_channel = 0; this_channel < local_channels; ++this_channel)
	{
		Global.channel = this_channel;
		channel_data[Global.channel].maximum_bits_to_remove = settings.static_maximum_bits_to_remove;
		channel_data[Global.channel].min_FFT_result.btr = channel_data[Global.channel].maximum_bits_to_remove;
		channel_data[Global.channel].min_FFT_result.analysis = 7;

		if (parameters.dynamic == 0)
		{
			process.dynamic_maximum_bits_to_remove = settings.static_maximum_bits_to_remove;
		}
		else
		{
			process.dynamic_maximum_bits_to_remove = int(std::max(0, int(BlockData.Channel_Log2_RMS[Global.channel] - settings.dynamic_minimum_bits_to_keep)));
		}

		if (process.dynamic_maximum_bits_to_remove < channel_data[Global.channel].maximum_bits_to_remove)
		{
			channel_data[Global.channel].maximum_bits_to_remove = process.dynamic_maximum_bits_to_remove;
			channel_data[Global.channel].min_FFT_result.btr = process.dynamic_maximum_bits_to_remove;
			channel_data[Global.channel].min_FFT_result.analysis = 8;
		}

		for (this_analysis_number = 1; this_analysis_number <= PRECALC_ANALYSES; ++this_analysis_number)
		{
			Global.analysis_number = this_analysis_number;
			Set_FFT_Data_Rec(Global.FFT, Global.analysis.bits[Global.analysis_number]);
			results.this_FFT.btr = 99;
			results.this_FFT.spreading = db_MaxDB;
			results.this_FFT.analysis = Global.analysis_number;

			if (((settings.FFT_analysis_switches & Global.FFT.length) > 0) && (Global.FFT.length <= Global.Codec_Block_size * 2))
			{
				Set_FFT_Variables();
				Zero_FFT_Raw_Results();
				process.Analyses_Performed = 0;

				for (this_analysis_block_number = 0; this_analysis_block_number <= process.analysis_blocks[Global.FFT.bit_length]; ++this_analysis_block_number)
				{
					FillFFT.actual_analysis_block_start = int(this_analysis_block_start + this_analysis_block_number * this_fft_underlap_length);
					FillFFT.actual_analysis_block_start = std::min(std::max(FillFFT.actual_analysis_block_start, min_neg_codec_block_start), max_pos_codec_block_start);

					if ((this_analysis_block_number == 0) && (FillFFT.actual_analysis_block_start == results.saved_FFT_results[Global.channel][Global.analysis_number].start))
					{
						if ((Global.FFT.length & settings.raw_result_short) > 0)
						{
							Fill_Last_from_History();
						}

						if ((Global.FFT.length & settings.raw_result_sgns) > 0)
						{
							Fill_Last_from_SGNS();
						}

						spreading.old_minimum = process.FFT_spreading[Global.analysis_number][Global.channel].old_minimum;
						spreading.new_minimum = process.FFT_spreading[Global.analysis_number][Global.channel].new_minimum;
						spreading.alt_average = process.FFT_spreading[Global.analysis_number][Global.channel].alt_average;
					}
					else if (FillFFT_Input_From_Data() == 0)
					{
						if ((Global.FFT.length & (settings.raw_result_short | settings.raw_result_sgns)) > 0)
						{
							Fill_Last_with_Zero();
						}

						spreading.old_minimum = db_MaxDB;
						spreading.new_minimum = db_MaxDB;
						spreading.alt_average = db_MaxDB;
					}
					else
					{
						if (FFTW_Initialised())
						{
							FFTW.Execute(FFTW.Plans[Global.FFT.bit_length]);
						}
						else
						{
							FFT_DIT_Real();
						}

						process.Post_Analysis[Global.analysis_number];
						Spreading_Function();
					}

					++ process.Analyses_Performed;
					Spreading_result = spreading.alt_average;
					Spreading_Used = 1;

					if (spreading.new_minimum < Spreading_result)
					{
						Spreading_result = spreading.new_minimum;
						Spreading_Used = 2;
					}

					if (spreading.old_minimum < Spreading_result)
					{
						Spreading_result = spreading.old_minimum;
						Spreading_Used = 3;
					}

					results.this_FFT.spreading = std::max(0.0, Spreading_result - FillFFT.Window_Function_RMS);
					results.this_FFT.btr = spreading.threshold_index[int(std::min(THRESHOLD_INDEX_SPREAD_MUL - 1.0, results.this_FFT.spreading) * THRESHOLD_INDEX_SPREAD)];
					results.this_FFT.start = FillFFT.actual_analysis_block_start - Global.Codec_Block_size;
					results.this_FFT.analysis = Global.analysis_number;

					if (parameters.spread != -1)
					{
						switch (Spreading_Used)
						{
							case 1:
								++ process.Alt_Ave_Used[Global.analysis_number];
								break;

							case 2:
								++ process.New_Min_Used_History[Global.analysis_number][process.new_min_bin];
								++ process.New_Min_Used[Global.analysis_number];
								break;

							case 3:
								++ process.Old_Min_Used_History[Global.analysis_number][process.old_min_bin];
								++ process.Old_Min_Used[Global.analysis_number];
								break;
						}

						if (results.this_FFT.btr > settings.static_maximum_bits_to_remove)
						{
							++ process.Over_Static[Global.analysis_number];
						}

						if (results.this_FFT.btr > process.dynamic_maximum_bits_to_remove)
						{
							++ process.Over_Dynamic[Global.analysis_number];
						}
					}

					if ((results.this_FFT.btr < channel_data[Global.channel].min_FFT_result.btr) || (channel_data[Global.channel].min_FFT_result.btr == -1))
					{
						channel_data[Global.channel].min_FFT_result = results.this_FFT;
					}
				}

				process.Analyses_Completed[Global.analysis_number] = process.Analyses_Completed[Global.analysis_number] + process.Analyses_Performed;

				if ((Global.FFT.length & settings.raw_result_short) > 0)
				{
					Fill_History_from_Last();
				}

				if ((Global.FFT.length & settings.raw_result_sgns) > 0)
				{
					Fill_SGNSFFT_from_Last();
				}

				results.saved_FFT_results[Global.channel][Global.analysis_number] = results.this_FFT;
				process.FFT_spreading[Global.analysis_number][Global.channel].old_minimum = spreading.old_minimum;
				process.FFT_spreading[Global.analysis_number][Global.channel].new_minimum = spreading.new_minimum;
				process.FFT_spreading[Global.analysis_number][Global.channel].alt_average = spreading.alt_average;
			}
		}

		if (parameters.adaptive_shaping != -1)
		{
			Make_Filter();
		}

		if (channel_data[Global.channel].min_FFT_result.btr < 0)
		{
			channel_data[Global.channel].min_FFT_result.btr = 0;
		}

		channel_data[Global.channel].bits_to_remove = channel_data[Global.channel].min_FFT_result.btr;

		if (Global.channel < Global.channels)
		{
			remove_bits_data.bits_to_remove = channel_data[Global.channel].bits_to_remove;
			Remove_Bits();
			channel_data[Global.channel].bits_to_remove = remove_bits_data.bits_removed;
			channel_data[Global.channel].clips = remove_bits_data.clips;
			codec_block_dependent_bits_to_remove = std::min(codec_block_dependent_bits_to_remove, channel_data[Global.channel].bits_to_remove);
		}
	}

	for (this_channel = 0; this_channel < Global.channels; ++this_channel)
	{
		Global.channel = this_channel;
		++ results.minima[ Global.channel ][ channel_data[Global.channel].min_FFT_result.analysis ];

		if (parameters.linkchannels || parameters.midside)
		{
			remove_bits_data.bits_to_remove = codec_block_dependent_bits_to_remove;

			if (channel_data[Global.channel].bits_to_remove != remove_bits_data.bits_to_remove)
			{
				if (Global.channel < Global.channels)
				{
					Remove_Bits();
				}

				channel_data[Global.channel].bits_to_remove = remove_bits_data.bits_removed;
				channel_data[Global.channel].clips = remove_bits_data.clips;
			}
		}

		channel_data[Global.channel].bits_lost = channel_data[Global.channel].min_FFT_result.btr - channel_data[Global.channel].bits_to_remove;
		++ Stats.bits_removed[ Global.channel ][ channel_data[Global.channel].bits_to_remove ];

		Global.bits_removed += channel_data[Global.channel].bits_to_remove;
		Global.bits_lost += channel_data[Global.channel].bits_lost;
		Global.clipped_samples += channel_data[Global.channel].clips;

		if (parameters.detail)
		{
			bit_removal_history[ (Global.blocks_processed - 1) * Global.channels + Global.channel ] = channel_data[Global.channel].bits_to_remove;
		}

		bits_removed_this_codec_block += OneOver[Global.channels] * channel_data[Global.channel].bits_to_remove;
	}

	//==========================================================================
	// Post analyse bit removed / correction audio data using short FFT.
	//==========================================================================
	if (parameters.postanalyse)
	{
		Set_FFT_Data_Rec(Global.FFT, history.FFT.bit_length);
		Set_FFT_Variables();

		for (this_channel = 0; this_channel < Global.channels; ++this_channel)
		{
			Global.channel = this_channel;

			//======================================================================
			// Create spectrum for BTRD-Codec-Block
			//======================================================================
			Fill_FFT_Proc = FillFFT_Input_From_BTRD;
			Add_to_LongDist_Proc = Add_To_BTRD_History;

			//======================================================================
			// Create spectrum for CORR-Codec-Block
			//======================================================================
			Process_Data_to_LongDist();
			Fill_FFT_Proc = FillFFT_Input_From_CORR;
			Add_to_LongDist_Proc = Add_To_CORR_History;
			Process_Data_to_LongDist();
		}
	}

	//==========================================================================
	// Post analyse DATA/BTRD/CORR audio data using Codec_Block_Length FFT.
	//==========================================================================
	if (parameters.longdist)
	{
		Set_FFT_Data_Rec(Global.FFT, LongDist.FFT.bit_length);
		Set_FFT_Variables();

		for (this_channel = 0; this_channel < Global.channels; ++this_channel)
		{
			Global.channel = this_channel;

			//======================================================================
			// Create spectrum for DATA-Codec-Block
			//======================================================================
			Fill_FFT_Proc = FillFFT_Input_From_Data;
			Add_to_LongDist_Proc = Add_to_DATA_LongDist;
			Process_Data_to_LongDist();

			//======================================================================
			// Create spectrum for BTRD-Codec-Block
			//======================================================================
			Fill_FFT_Proc = FillFFT_Input_From_BTRD;
			Add_to_LongDist_Proc = Add_to_BTRD_LongDist;
			Process_Data_to_LongDist();

			//======================================================================
			// Create spectrum for CORR-Codec-Block
			//======================================================================
			Fill_FFT_Proc = FillFFT_Input_From_CORR;
			Add_to_LongDist_Proc = Add_to_CORR_LongDist;
			Process_Data_to_LongDist();
		}
	}

	//==========================================================================
	// Create DATA, BTRD and CORR sample value histogram
	//==========================================================================
	if (parameters.histogram)
	{
		Add_to_Histogram();
	}

	//==========================================================================
	// Store BTRD and CORR results, if necessary, for use next codec-block.
	//==========================================================================
	if (parameters.postanalyse || parameters.longdist)
	{
		Fill_Audio_Safes();
	}

	Global.samples_processed = (Global.blocks_processed - 1) * Global.Codec_Block_size + BlockData.Size.This;

	Process_Output();

	//==========================================================================
	// Create histogram of LSBs and MSBs on a sample / block basis.
	//==========================================================================
	if (parameters.blockdist || parameters.sampledist)
	{
		lsb_analysis();
	}
}


void nProcess_Init()
{
	int sa_i, sa_j;
	int total_overlap_length;

	for (sa_j = 1; sa_j <= PRECALC_ANALYSES; ++sa_j)
	{
		for (sa_i = 0; sa_i < MAX_FFT_LENGTH_HALF; ++sa_i)
		{
			process.Old_Min_Used_History[sa_j][sa_i] = 0;
			process.New_Min_Used_History[sa_j][sa_i] = 0;
		}

		process.Old_Min_Used[sa_j] = 0;
		process.New_Min_Used[sa_j] = 0;
		process.Alt_Ave_Used[sa_j] = 0;
		process.Over_Dynamic[sa_j] = 0;
		process.Over_Static[sa_j] = 0;
		process.Post_Analysis[sa_j] = Post_Process_FFT_Results_No_Raw;
	}

	if (settings.raw_result_short != 0)
	{
		process.Post_Analysis[2] = Post_Process_FFT_Results;
	}

	if (settings.raw_result_sgns != 0)
	{
		process.Post_Analysis[6] = Post_Process_FFT_Results;
	}

	for (sa_i = 1; sa_i <= MAX_FFT_BIT_LENGTH; ++sa_i)
	{
		Set_FFT_Data_Rec(Global.FFT, sa_i);

		if (FFTW_Initialised())
		{
			if (((settings.FFT_analysis_switches & Global.FFT.length) > 0) || (parameters.longdist && (Global.FFT.bit_length == LongDist.FFT.bit_length)))
			{
				FFTW.Plans[Global.FFT.bit_length] = FFTW.Plan_DFT_r2c_1d(Global.FFT.length, &FFT_Array.DReal[0], &FFT_Array.DReal[0], 0x69);
			}
		}

		//========================================================================
		// Calculate actual analysis_time values for FFT_lengths
		//========================================================================
		process.end_overlap_length[Global.FFT.bit_length] = std::min(Global.FFT.length / 2, Global.Codec_Block_size);
		process.actual_analysis_blocks_start[Global.FFT.bit_length] = -process.end_overlap_length[Global.FFT.bit_length];

		total_overlap_length = Global.Codec_Block_size + process.end_overlap_length[Global.FFT.bit_length] * 2 - Global.FFT.length;
		process.FFT_underlap_length[Global.FFT.bit_length] = std::max(1, Global.FFT.length / std::max(2, parameters.underlap));

		process.analysis_blocks[Global.FFT.bit_length] = std::max(0, round(total_overlap_length / process.FFT_underlap_length[Global.FFT.bit_length]));

		process.FFT_underlap_length[Global.FFT.bit_length] = total_overlap_length / std::max(1, process.analysis_blocks[Global.FFT.bit_length]);
	}
}
// initialization: todo
// finalization: todo Write ( Erroutput, UnitName, CRLF );
