//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

//#include <cmath>
#include "nCore.h"
#include "nMaths.h"
#include "nSpreading.h"

static const char UnitName [] = "nSpreading";

// Globals
spreading_type spreading;
double Skewing_Gain[(MAX_FFT_LENGTH / 2) + 2 ];


void Spreading_Function()
{
	int sc_i, sc_j, sc_w, sc_f;
	double new_value, old_value, alt_average;
	spreading.new_minimum = db_MaxDB;
	spreading.old_minimum = db_MaxDB;
	alt_average = 0;

	for (sc_i = spreading.Bins.Lo[Global.FFT.bit_length]; sc_i <= spreading.Bins.hi[Global.FFT.bit_length]; ++sc_i)
	{
		/*with spreading.averages_ptr[Global.analysis_number]^[sc_i] do*/
		Spreading_rec& rec = spreading.averages_ptr[Global.analysis_number][sc_i];
		{
			sc_w = rec.width;
			sc_f = rec.fractint;
		}
		alt_average += FFT_skew_result[sc_i];
		old_value = 0;

		for (sc_j = 0; sc_j < sc_w; ++sc_j)
		{
			old_value += FFT_skew_result[sc_i - sc_j];
		}

		old_value *= OneOver[sc_w];
		new_value = ((FFT_skew_result[sc_i - 1] + FFT_skew_result[sc_i + 1]) * spreading.Widths[sc_f] + FFT_skew_result[sc_i]) * spreading.divisors[sc_f];

		if (spreading.old_minimum > old_value)
		{
			spreading.old_minimum = old_value;
			process.old_min_bin = sc_i;
		}

		if (spreading.new_minimum > new_value)
		{
			spreading.new_minimum = new_value;
			process.new_min_bin = sc_i;
		}
	}

	spreading.alt_average = (nlog2(alt_average * spreading.Bins.Recip[Global.FFT.bit_length]) + Global.FFT.threshold_shift) * db_lg2x20 + settings.noise_threshold_shift_average;
	spreading.old_minimum = (nlog2(spreading.old_minimum) + Global.FFT.threshold_shift) * db_lg2x20 + settings.noise_threshold_shift_minimum;
	spreading.new_minimum = (nlog2(spreading.new_minimum) + Global.FFT.threshold_shift) * db_lg2x20 + settings.noise_threshold_shift_minimum;
}


void Spreading_Memory_Allocate()
{
	int sa_i;

	for (sa_i = 1; sa_i <= PRECALC_ANALYSES; ++sa_i)
	{
		spreading.averages_ptr[sa_i] = new Spreading_rec[ Global.analysis.length[sa_i] * 2 + 4 ];
	}
}


void Skewing_Curve_Init()
{
	double sa_lfb, sa_mfb, sa_tfb, sa_amp;
	int sa_i;
	sa_lfb = log10_2 * nlog2(LOWER_FREQ_LIMIT);
	sa_mfb = log10_2 * nlog2(MID_FREQ_LIMIT);

	if (parameters.dccorrect)
	{
		sa_amp = SKEWING_AMPLITUDE * SINGLE_HALF;
	}
	else
	{
		sa_amp = SKEWING_AMPLITUDE;
	}

	for (sa_i = 0; sa_i <= PowersOf.TwoM1[MAX_FFT_BIT_LENGTH - 1] + 1; ++sa_i)
	{
		sa_tfb = log10_2 * nlog2((sa_i * PowersOf.Two[- MAX_FFT_BIT_LENGTH] * Global.sample_rate));

		if (sa_tfb <= sa_mfb)
		{
			Skewing_Gain[sa_i] = npower2((std::pow(std::sin(HalfPi * std::max(0.0, sa_tfb - sa_lfb) / (sa_mfb - sa_lfb)), 0.75) - 1) * sa_amp * OneOver[20] * log2_10);
		}
		else
		{
			Skewing_Gain[sa_i] = 1;
		}
	}

	Skewing_Gain[0] = 0;
}


void Spreading_Widths_Init()
{
	int sa_i;

	for (sa_i = 0; sa_i <= SPREADING_STEPS; ++sa_i)
	{
		spreading.Widths[sa_i] = float(sa_i * SPREADING_STEPS_RECIP);
		spreading.divisors[sa_i] = 1.0f / (1.0f + (2.0f * spreading.Widths[sa_i]));
	}
}


void Threshold_Index_Init()
{
	int last_filled;
	int this_bit_to_remove;
	double this_reference_threshold;
	last_filled = 0;

	for (this_bit_to_remove = 1; this_bit_to_remove <= BITS_TO_CALCULATE; ++this_bit_to_remove)
	{
		this_reference_threshold = -2.0 + this_bit_to_remove + Global.Codec_Block_bits * SINGLE_HALF
		                         + nlog2(1 + PowersOf.Two[1 - this_bit_to_remove * 2]) * SINGLE_HALF
		                         + nlog2(1 + PowersOf.Two[- 1 - Global.Codec_Block_bits]) * 0.30769230769230769230769230769231
		                         - 0.20751874963942190927313052802609
		                         - 0.13444987323232079693278710922535;

		this_reference_threshold *= db_lg2x20 * THRESHOLD_INDEX_SPREAD;

		while (last_filled < this_reference_threshold)
		{
			spreading.threshold_index[last_filled] = this_bit_to_remove - 1;
			++ last_filled;
		}
	}

	while (last_filled < THRESHOLD_INDEX_SPREAD_MAX)
	{
		spreading.threshold_index[last_filled] = BITS_TO_CALCULATE - 1;
		++ last_filled;
	}
}


void Spreading_Memory_Free()
{
	int sa_i;

	for (sa_i = 1; sa_i <= PRECALC_ANALYSES; ++sa_i)
	{
		delete[] spreading.averages_ptr[sa_i];
	}
}


void nSpreading_Init()
{
	int this_analysis_number;
	int sa_i, sa_j;
	Spreading_Memory_Allocate();
	Spreading_Widths_Init();
	Skewing_Curve_Init();
	Threshold_Index_Init();

	for (sa_i = 1; sa_i <= MAX_FFT_BIT_LENGTH; ++sa_i)
	{
		spreading.Bins.Lo[sa_i] = std::max(1, round(double(Global.lower_freq_limit) / Global.sample_rate * PowersOf.Two[sa_i]) - 1);
		spreading.Bins.hi[sa_i] = std::max(1, round(double(Global.upper_freq_limit) / Global.sample_rate * PowersOf.Two[sa_i]) - 1);
		spreading.Bins.Recip[sa_i] = OneOver[spreading.Bins.hi[sa_i] - spreading.Bins.Lo[sa_i] + 1];
		spreading.Bins.mid[sa_i] = std::max(1, round(double(MID_FREQ_LIMIT) / Global.sample_rate * PowersOf.Two[sa_i]) - 1);
	}

	for (this_analysis_number = 1; this_analysis_number <= PRECALC_ANALYSES; ++this_analysis_number)
	{
		Global.analysis_number = this_analysis_number;
		Set_FFT_Data_Rec(Global.FFT, history.FFT.bit_length - 2 + Global.analysis_number);

		for (sa_i = 0; sa_i <= SPREAD_ZONES + 1; ++sa_i)
		{
			spreading.Frequency_Bins[Global.analysis_number][sa_i] = std::min(Global.FFT.length / 2 - 1, std::max(1, round(Frequency_Limits[sa_i] / Global.sample_rate * Global.FFT.length)));

			if (spreading.Frequency_Bins[Global.analysis_number][sa_i] > Global.FFT.length / 2 - 1)
			{
				lossyWAVerror(std::string("frequency too high : ") + floattostrf(Frequency_Limits[sa_i]* OneOver[1000], ffFixed, 99, 3) + "kHz", 0x21);
			}
		}

		for (sa_j = SPREAD_ZONES; sa_j >= 0; --sa_j)
		{
			for (sa_i = spreading.Frequency_Bins[Global.analysis_number][sa_j]; sa_i <= spreading.Frequency_Bins[Global.analysis_number][sa_j + 1]; sa_i ++)
			{
				spreading.averages_ptr[Global.analysis_number][sa_i].width = std::min(unsigned short(sa_i), unsigned short(SPREADING_FUNCTION_ARRAY[Global.analysis_number][sa_j]));
			}
		}

		for (sa_i = 0; sa_i <= spreading.Bins.mid[Global.FFT.bit_length]; ++sa_i)
		{
			spreading.averages_ptr[Global.analysis_number][sa_i].fractint = 0;
		}

		for (sa_i = spreading.Bins.mid[Global.FFT.bit_length]; sa_i <= Global.FFT.length_half_m1; ++sa_i)
		{
			spreading.averages_ptr[Global.analysis_number][sa_i].fractint = unsigned short(SPREADING_STEPS * std::max(0.0, std::min(double(UPPER_FREQ_LIMIT - MID_FREQ_LIMIT), sa_i * double(Global.sample_rate) / Global.FFT.length - MID_FREQ_LIMIT)) / (UPPER_FREQ_LIMIT - MID_FREQ_LIMIT) * SPREADING_FUNCTION_WIDTHS[Global.analysis_number]);
		}
	}
}


static struct Init
{
	Init() {
		// initialization: todo
	}
	~Init() {
		// finalization: todo Write ( Erroutput, UnitName, CRLF );
		Spreading_Memory_Free();
	}
} init;
