//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

static const int SKEWING_AMPLITUDE = 36;
static const int SPREADING_REC_SIZE = 4;
static const int SPREADING_STEPS = 8190;
static const double SPREADING_STEPS_RECIP = 1.0 / SPREADING_STEPS;

static const int THRESHOLD_INDEX_SPREAD = 64;
static const int THRESHOLD_INDEX_SPREAD_MUL = 256;
static const int THRESHOLD_INDEX_SPREAD_MAX = THRESHOLD_INDEX_SPREAD_MUL * THRESHOLD_INDEX_SPREAD;

//============================================================================
// Frequency range over which to check for lowest amplitude signal
//============================================================================
static const int SPREAD_ZONES = 7;
static double Frequency_Limits[ SPREAD_ZONES + 2 ] = { 20, 1378.125, 3445.3125, 5512.5, 8268.75, 10335.9375, 12403.125, 14470.3125, 16000 };

static unsigned char SPREADING_FUNCTION_ARRAY[ PRECALC_ANALYSES + 1 ][ SPREAD_ZONES + 1 ] =
{
	{ 0 }, // index 0 not used // TY
	{ 1, 2, 2, 2, 2, 2, 2, 2 },
	{ 1, 2, 2, 2, 2, 2, 2, 2 },
	{ 1, 2, 2, 2, 2, 2, 2, 3 },
	{ 1, 2, 2, 2, 2, 2, 2, 3 },
	{ 1, 2, 2, 2, 2, 2, 2, 4 },
	{ 1, 1, 2, 2, 2, 2, 3, 4 }
};

static const double SPREADING_FUNCTION_WIDTHS[ PRECALC_ANALYSES + 1 ] =
{
	0, // index 0 not used // TY
	0.25,
	0.17677669529663688110021109052621,
	0.250,
	0.35355339059327376220042218105242,
	0.5,
	0.70710678118654752440084436210485
};


struct Spreading_rec
{
	unsigned short width;
	unsigned short fractint;
};


typedef Spreading_rec sprec_array[ 1 ];
typedef Spreading_rec* sprec_array_ptr;


// Vars

extern struct spreading_type
{
	float Widths[ SPREADING_STEPS + 2 ];
	float divisors[ SPREADING_STEPS + 2 ];

	double alt_average;
	double old_minimum;
	double new_minimum;

	sprec_array_ptr averages_ptr[ PRECALC_ANALYSES + 1 ]; // base 1

	//============================================================================
	// For each FFT bit length store bits_to_remove with respect to calculated minimum dB of FFT result.
	//============================================================================
	unsigned char threshold_index[ THRESHOLD_INDEX_SPREAD_MAX ];

	//============================================================================
	// FFT Bin associated with relevant frequencies for each FFT length.
	//============================================================================
	int Frequency_Bins[ PRECALC_ANALYSES + 1 ][ SPREAD_ZONES + 2 ];

	//============================================================================
	// FFT bins for defining calculation limits.Spreading.
	struct
	{
		int Lo[ MAX_FFT_BIT_LENGTH + 1 ];
		int hi[ MAX_FFT_BIT_LENGTH + 1 ];
		int mid[ MAX_FFT_BIT_LENGTH + 1 ];
		double Recip[ MAX_FFT_BIT_LENGTH + 1 ];
	}
	Bins;
}
spreading;

//============================================================================
// skewing function low frequency bin bin attenuation lookup table.
//============================================================================
extern double Skewing_Gain[(MAX_FFT_LENGTH / 2) + 2 ];

void nSpreading_Init();
void Spreading_Function();